/*
 * Copyright (C) 2004 Rutgers University
 *
 * Permission is hereby granted to distribute this software for
 * non-commercial research purposes, provided that this copyright
 * notice is included with any such distribution.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE
 * SOFTWARE IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU
 * ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
 *
 * Author: John Asmuth jasmuth@cs.rutgers.edu
 */

#include <iostream>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <vector>
#include <stdio.h>

using namespace std;

char* color_table[] = {"red", "blue", "green", "cyan", "magenta"};
int max_colors = 5;

int random_int(int range)
{
  return int(range*drand48())%range;
}

char* random_color(int max)
{
  return color_table[random_int(max)];
}

int getBlockCount(float prob)
{

  int count = 1;
  while (drand48() > prob)
    count++;
  return count;
}

int getColorCount(float prob)
{
  return 3;
  /*
  int count = 1;
  while (drand48() > prob)
    count++;
  return count;
  */
}

bool preconditions = false;
bool metric = false;
char* problem_name = "blocks-problem";
char* domain_name = "blocks-domain";
int color_count;
int block_count;
int preward = 1;
int dreward = 1;
int greward = 0;

int bc_count;
char* *bc_vec;

void printDomain();
void printActions();
void printProblem();
void printInitStack();
void printInitColors();
void printGoal();
void printStackNC();
void printRecursiveStack(int block_index, int tower_index);

void printDomain()
{
  cout << "(define (domain " << domain_name << ")" << endl
       << " (:requirements :adl :probabilistic-effects :fluents";
	   if((preward!=1) || (dreward!=1)){
		cout << " :rewards";
	   }
  cout << ")" << endl
       << " (:types block table)" << endl
       << " (:constants table - table)" << endl;

  cout << " (:predicates" << endl;
  for (int i=0; i<color_count; i++)
    cout << "  (is-" << color_table[i] << " ?block - block)" << endl;
  cout << "  (holding ?block - block)" << endl
       << "  (on-top-of ?block - block ?obj)" << endl
       << " )" << endl;

  printActions();

  cout << ")" << endl;
}

void printActions()
{
  cout << " (:action pick-up-block-from" << endl
       << "  :parameters (?top - block ?bottom)" << endl;

  if (preconditions) {
    cout << "  :precondition (and (not (= ?top ?bottom))" << endl
	 << "                     (forall (?b - block) (not (holding ?b)))" << endl
	 << "                     (on-top-of ?top ?bottom)" << endl
	 << "                     (forall (?b - block) (not (on-top-of ?b ?top))))" << endl;
  }
  cout << "  :effect ";

  if (!preconditions) {

    cout << "(when (and (not (= ?top ?bottom))" << endl
	 << "                     (forall (?b - block) (not (holding ?b)))" << endl
	 << "                             (on-top-of ?top ?bottom)" << endl
	 << "                     (forall (?b - block) (not (on-top-of ?b ?top))))" << endl;
  }

  cout << "                (and ";
  if(preward != 1){
    cout << "(decrease (reward) " << (0-preward) << ")" << endl;
    cout << "               ";
  }
  char* lead = "";
  if (!preconditions) {
    lead = "      ";
  }
  cout << "(probabilistic 0.75 (and (holding ?top)" << endl
       << lead << "                                        (not (on-top-of ?top ?bottom)))" << endl
       << lead << "                              0.25 (when (not (= ?bottom table))" << endl
       << lead << "                                         (and (not (on-top-of ?top ?bottom))" << endl
       << lead << "                                              (on-top-of ?top table)))))" << endl;
  if (!preconditions) {
    cout << "          )" << endl;
  }
  cout << " )" << endl;

  cout << " (:action put-down-block-on" << endl
       << "  :parameters (?top - block ?bottom)" << endl;
  if (preconditions) {
    cout << "  :precondition (and (not (= ?top ?bottom))" << endl
	 << "                     (holding ?top)" << endl
	 << "                     (or (= ?bottom table)" << endl
	 << "                         (forall (?b - block) (not (on-top-of ?b ?bottom)))))" << endl;
  }
  cout << "  :effect ";
  if (!preconditions) {
    cout << "(when (and (not (= ?top ?bottom))" << endl
	 << "                     (holding ?top)" << endl
	 << "                     (or (= ?bottom table)" << endl
	 << "                     (forall (?b - block) (not (on-top-of ?b ?bottom)))))" << endl;
  }

  cout << "                (and ";
  if(dreward != 1) {
    cout << "(decrease (reward) " << (0-dreward) << ")" << endl
	 << "               ";
  }
  cout << "(not (holding ?top))" << endl;
  lead = "";
  if (!preconditions) {
    lead = "      ";
  }
  cout << lead << "               (probabilistic 0.75 (on-top-of ?top ?bottom)" << endl;
  cout << lead << "                              0.25 (on-top-of ?top table)))" << endl;
  if (!preconditions) {
    cout << "          )" << endl;
  }
  cout << " )" << endl;
}

void printProblem()
{
  cout << "(define (problem " << problem_name << ")" << endl
       << " (:domain " << domain_name << ")" << endl
       << " (:objects";
  for (int i=0; i<block_count; i++)
    cout << " block" << i;
  cout << " - block)" << endl;
  cout << " (:init" << endl;
  printInitStack();
  printInitColors();
  cout << " )" << endl;
  printGoal();
  cout << ")" << endl;
}

void printInitStack()
{
  for (int block_index = 0; block_index < block_count;) {
    int tower_size = random_int(block_count-block_index)+1;
    for (int i=0; i<tower_size-2; i++) {
      cout << "  (on-top-of block" << block_index
           << " block" << block_index+1 << ")" << endl;
      block_index++;
    }
    cout << "  (on-top-of block" << block_index << " table)" << endl;
    block_index++;
  }
}

void printInitColors()
{
  if (color_count == 0)
    return;
  bc_count = block_count;
  bc_vec = new char*[bc_count];
  for (int i=0; i<block_count; i++)
    bc_vec[i] = random_color(color_count);
  for (int i=0; i<block_count; i++)
    cout << "  (is-" << bc_vec[i] << " block" << i << ")" << endl;
}

// (exists (?f1) (and (is-red ?f1) next)

void printGoal()
{
  cout << " (:goal" << endl
       << "  (and" << endl;
  if (color_count)
    printRecursiveStack(0, random_int(block_count)+1);
  else
    printStackNC();
  cout << endl
       << "  )" << endl
       << " )" << endl;
  if (greward > 0) {
    cout << " (:goal-reward " << greward << ")" << endl;
  }
  if (metric && ((preward!=1) || (dreward!=1))) {
    cout << " (:metric maximize (reward))" << endl;
  }
}

void printStackNC()
{
  vector<int> blockvec;
  blockvec.push_back(0);
  for (int i=1; i<block_count; i++) {
    int l=blockvec.size();
    vector<int>::iterator itr=blockvec.begin();
    int insert_index=random_int(l);
    for (;insert_index>=0; insert_index--)
      itr++;
    blockvec.insert(itr, i);
  }

  for (int tower_index=0; tower_index<block_count;) {
    int tower_size = random_int(block_count-tower_index)+1;
    for (int cur=tower_index; cur<tower_index+tower_size; cur++) {
      if (cur == tower_index+tower_size-1) {
	cout << "   (on-top-of block" << blockvec[cur] << " table)" << endl;
      }
      else {
	cout << "   (on-top-of block" << blockvec[cur]
	     << " block" << blockvec[cur+1] << ")" << endl;
      }
    }
    tower_index += tower_size;
  }
}

void printRecursiveStack(int block_index, int tower_index)
{
  if (block_index == block_count) {
    cout << "     (on-top-of ?fb" << block_index-1 << " table)" << endl;
    return;
  }

  cout << "   (exists (?fb" << block_index << " - block)" << endl
       << "    (and" << endl;

  for (int i=0; i<block_index; i++)
    cout << "     (not (= ?fb" << i << " ?fb" << block_index << "))" << endl;

  int color_index = random_int(bc_count);
  char *color = bc_vec[color_index];
  cout << "     (is-" << color << " ?fb" << block_index << ")" << endl;
  for (int i=color_index; i<bc_count-1; i++)
    bc_vec[i] = bc_vec[i+1];
  bc_count--;

  if (block_index) {
    cout << "     (on-top-of ?fb" << block_index-1;
    if (tower_index)
      cout << " ?fb" << block_index;
    else
      cout << " table";
    cout << ")" << endl;
  }

  if (!tower_index)
    tower_index = random_int(block_count-block_index)+2;

  printRecursiveStack(block_index+1, tower_index-1);
  cout << "))";
}

void usage(char *name)
{
  char* lead = strdup(name);
  for (char* s = lead; *s; s++)
    *s = ' ';
  cout << "usage: " << name << " [-b block-count]" << endl
       << "       " << lead << " [-c color-count]" << endl
       << "       " << lead << " [-rp pickup-cost]" << endl
       << "       " << lead << " [-rd putdown-cost]" << endl
       << "       " << lead << " [-s seed]" << endl
       << "       " << lead << " [-pn problem-name]" << endl
       << "       " << lead << " [-dn domain-name]" << endl
       << "       " << lead << " [-gr goal reward]" << endl
       << "       " << lead << " [-mr (maximize reward)]" << endl
       << "       " << lead << " [-pc (preconditions)]" << endl;
}

int main(int argc, char **argv)
{
  long seed = time(0);

  color_count = -1;
  block_count = -1;

  for (int i=1; i<argc; i++) {
    if (!strcmp(argv[i], "-b")) {
      if (i==argc-1) {
	usage(*argv);
	return 1;
      }
      block_count = atoi(argv[++i]);
    }
    if(!(strcmp(argv[i], "-gr"))){
      if(i == argc - 1){
	usage(*argv);
	return 1;
      }
      greward = atoi(argv[++i]);
      if(greward <= 0){
	usage(*argv);
	return 1;
      }
    }
    if (!strcmp(argv[i], "-c")) {
      if (i==argc-1) {
	usage(*argv);
	return 1;
      }
      color_count = atoi(argv[++i]);
      if(color_count > 5){
	usage(*argv);
	return 1;
      }
    }
    if (!strcmp(argv[i], "-s")) {
      if (i==argc-1) {
	usage(*argv);
	return 1;
      }
      seed = atol(argv[++i]);
    }
    if (!strcmp(argv[i], "-pc")) {
      preconditions = true;
    }
    if (!strcmp(argv[i], "-mr")) {
      metric = true;
    }
    if (!strcmp(argv[i], "-pn")) {
      if(i==argc-1) {
	usage(*argv);
	return 1;
      }
      problem_name = argv[++i];
    }
    if (!strcmp(argv[i], "-dn")) {
      if(i==argc-1) {
	usage(*argv);
	return 1;
      }
      domain_name = argv[++i];
    }
    if (!strcmp(argv[i],"-rp")) {
      if(i==argc-1) {
	usage(*argv);
	return 1;
      }
      preward = atoi(argv[++i]);
      if(preward >= 0) {
	usage(*argv);
	return 1;
      }
    }
    if (!strcmp(argv[i],"-rd")) {
      if(i==argc-1) {
	usage(*argv);
	return 1;
      }
      dreward = atoi(argv[++i]);
      if(dreward >= 0) {
	usage(*argv);
	return 1;
      }
    }
    if (!strcmp(argv[i], "-?")) {
      usage(*argv);
      return 0;
    }
  }

  srand48(seed);

  if (color_count == -1)
    color_count = getColorCount(0.5);
  if (block_count == -1)
    block_count = getBlockCount(0.1);

  cout << ";; Generated by blocksworld generator" << endl
       << ";; http://www.cs.rutgers.edu/~jasmuth/blocksworld.tar.gz" << endl
       << ";; by John Asmuth (jasmuth@cs.rutgers.edu)" << endl
       << ";; and Dave Weissman" << endl << endl;

  printDomain();
  printProblem();

  return 0;

}

